# -*- encoding: utf-8 -*-
"""
hio.core.multidoing Module

Provides multiprocessing support



"""
import os
import time
import logging
import multiprocessing as mp

from collections import deque, namedtuple
from dataclasses import dataclass, astuple, asdict, field


from . import tyming
from .doing import Doist, Doer
from .. import hioing
from .. import help
from ..help import timing, helping
from ..help.helping import RawDom
from ..help import ogling


logger = help.ogler.getLogger()


@dataclass
class DoistDom(RawDom):
    """Configuration dataclass to child process Doist parameters for MultiDoer

    Attributes:
        name (str): child doist identifier for resources.
        tyme (float): child doist start tyme
        tock (float | None): child doist tock, tyme lag between runs
        real (bool): child doist True means run in real time, tyme is real time
        limit (float | None): child doist tyme limit. None means run forever
        doers (list[Doers]): child doist List of Doers
    """
    name: str ='child'  # unique identifier of child process and associated resources
    tyme: float = 0.0    # child doist start tyme
    tock: float | None = None  # child doist tyme lag between runs, None means ASAP
    real: bool = True  # child doist True means run in real time, tyme is real time
    limit: float | None = None  # child doist tyme limit. None mean run forever
    doers: list = field(default_factory=list)  # child doist list of doers

    def __iter__(self):
        return iter(asdict(self))



class MultiDoer(Doer):
    """
    MultiDoer spawns multiprocesses with Doists

    See Doer for inherited attributes, properties, and methods.

    Inherited Attributes:
        done (bool): completion state:
                     True means completed fully. Otherwise incomplete.
                     Incompletion value may be None or False.
        opts (dict): schedulaer injects options from .opts into its .do generator
                     function as **opts parameter.


    Inherited Properties:
        tyme (float): is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap


    Inherited Methods:
        __call__()  makes instance callable as generator function returning generator
        do() generator method that returns generator
        enter() is enter context action method
        recur() recur context action method or generator method
        clean() clean context action method
        exit() exit context method
        close() close context method
        abort() abort context method
        wind()  injects ._tymth dependency from associated Tymist to get its .tyme


    Attributes:
        ctx: (mp.context.SpawnContext | None): context under which to spawn processes
        loads: (list[dict]): each expanded to kwargs used to spinup a
                             child process' doist
        tots (dict): values are child Process instances keyed by name
        count (int): iteration count

    Properties:


    """

    def __init__(self, loads=None, temp=False, **kwa):
        """Initialize instance.


        Inherited Parameters:
            tymth (closure): injected function wrapper closure returned by
                Tymist.tymen() instance. Calling tymth() returns associated
                Tymist.tyme.
            tock (float): seconds initial value of .tock
            opts (dict): injected options into its .do generator by scheduler

        Parameters:
            loads (list[dict]): each expanded to kwargs used to spinup a
                                 child process' doist
            temp (bool): True means logger or other resources in spinup uses temp
                         False other wise

        """
        super(MultiDoer, self).__init__(**kwa)
        self.loads = loads if loads is not None else []
        self.temp = temp
        self.tots = {}
        self.ctx = mp.get_context('spawn')
        self.count = None


    def enter(self, *, temp=None):
        """Do 'enter' context actions.
        Start processes with config from .tots
        Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary local file resources if any
                                None means ignore parameter value use self.temp

        Doist or DoDoer winds its doers on enter

        """

        self.count = 0
        for load in self.loads:
            tot = self.ctx.Process(name=load["name"],
                                     target=self.spinup,
                                     kwargs=load)
            self.tots[tot.name] = tot
            tot.start()


    def recur(self, tyme):
        """"""
        self.count += 1

        if not self.ctx.active_children():
            return True   # complete
        return False  # incomplete recur again

    def exit(self):
        """"""
        self.count += 1


    def close(self):
        """"""
        self.count += 1


    def abort(self, ex):
        """"""
        self.count += 1


    @staticmethod
    def spinup(name, tyme, tock, real, limit, doers, temp):
        """Process target function to make and run doist after child subprocess has
        been started.

        Paramters:
            name (str): unique identifier for child doist and its resources
            tyme (float): child doist start tyme
            tock (float | None): child doist tyme lag between runs, None means ASAP
            real (bool): child doist runtime
                         True means run in real time, tyme is real time
                         False means run in faster than real time, tyme is relative
            limit (float | None): child doist tyme limit. None mean run forever
            doers (list): child doist list of doers. Do not assign tymist to doers
                          when creating.
            temp (bool): use temp resources such as file pathing

        Doist must be built after process started so local tymth closure is created
        inside subprocess so that when doist winds the tyme for its doers
        the tymth closure is locally sourced.
        """
        ogler = ogling.initOgler(prefix='hio_multi', name="hio_multi", level=logging.DEBUG,
                                              temp=temp, reopen=True, clear=True)
        logger = ogler.getLogger()

        logger.info("Child Starting: name=%s, ppid=%d, pid=%s, module=%s.",
                                name, os.getppid(), os.getpid(), __name__)
        time.sleep(0.01)

        doist = Doist(name=name, tock=tock, real=real, limit=limit, doers=doers)
        doist.do()

        logger.info("Child Ended: name=%s, ppid=%d, pid=%s, module=%s.",
                             name, os.getppid(), os.getpid(), __name__)





# -*- encoding: utf-8 -*-
"""
hio.core.uxd.peermemoing Module
"""
from contextlib import contextmanager

from ... import help
from ... import hioing
from ...base import doing
from ..uxd import Peer
from ..memo import Memoer, GramDex

logger = help.ogler.getLogger()


class PeerMemoer(Peer, Memoer):
    """Class for sending memograms over UXD transport
    Mixin base classes Peer and Memoer to attain memogram over uxd transport.


    Inherited Class Attributes:
        MaxGramSize (int): absolute max gram size on tx with overhead
        See memoing.Memoer Class
        See Peer Class

    Inherited Attributes:
        See memoing.Memoer Class
        See Peer Class

    Class Attributes:


    Attributes:


    """


    def __init__(self, *, bc=4, **kwa):
        """Initialization method for instance.

        Inherited Parameters:
            bc (int | None): count of transport buffers of MaxGramSize

            See memoing.Memoer for other inherited paramters
            See Peer for other inherited paramters


        Parameters:

        """
        super(PeerMemoer, self).__init__(bc=bc, **kwa)



@contextmanager
def openPM(cls=None, name="test", temp=True, reopen=True, clear=True,
             filed=False, extensioned=True, **kwa):
    """
    Wrapper to create and open UXD PeerMemoer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of PeerMemoer peer.
                    Enables management of transport by name.
                    Provides unique path part so can have many peers each at
                    different paths but in same directory.
        temp (bool): True means open in temporary directory, clear on close
                     Otherwise open in persistent directory, do not clear on close
        reopen (bool): True (re)open with this init
                       False not (re)open with this init but later (default)
        clear (bool): True means remove directory upon close when reopening
                      False means do not remove directory upon close when reopening
        filed (bool): True means .path is file path not directory path
                      False means .path is directiory path not file path
        extensioned (bool): When not filed:
                            True means ensure .path ends with fext
                            False means do not ensure .path ends with fext

    See filing.Filer and uxding.Peer for other keyword parameter passthroughs

    Usage:
        with openPM() as peer:
            peer.receive()

        with openPM(cls=PeerMemoerSubclass) as peer:
            peer.receive()

    """
    peer = None
    if cls is None:
        cls = PeerMemoer
    try:
        peer = cls(name=name, temp=temp, reopen=reopen, clear=clear,
                   filed=filed, extensioned=extensioned, **kwa)

        yield peer

    finally:
        if peer:
            peer.close(clear=peer.temp or clear)



class PeerMemoerDoer(doing.Doer):
    """PeerMemoerDoer Doer for reliable UXD transport.
    Does not require retry tymers.

    See Doer for inherited attributes, properties, and methods.
    To test in WingIde must configure Debug I/O to use external console

    Attributes:
       .peer (PeerMemoerDoer): underlying transport instance subclass of Memoer

    """

    def __init__(self, peer, **kwa):
        """Initialize instance.

        Parameters:
           peer (Peer): is Memoer Subclass instance
        """
        super(PeerMemoerDoer, self).__init__(**kwa)
        self.peer = peer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value use self.temp

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        self.peer.service()


    def exit(self):
        """"""
        self.peer.close(clear=True)

